# SPDX-License-Identifier: Apache-2.0

import random
import struct

import numpy as np  # type: ignore

from onnx import helper, defs, numpy_helper, checker
from onnx import AttributeProto, TensorProto, GraphProto, ModelProto, OptionalProto, TypeProto, SequenceProto
from typing import Any, List, Tuple

import unittest


class TestHelperAttributeFunctions(unittest.TestCase):

    def test_attr_float(self) -> None:
        # float
        attr = helper.make_attribute("float", 1.)
        self.assertEqual(attr.name, "float")
        self.assertEqual(attr.f, 1.)
        checker.check_attribute(attr)
        # float with scientific
        attr = helper.make_attribute("float", 1e10)
        self.assertEqual(attr.name, "float")
        self.assertEqual(attr.f, 1e10)
        checker.check_attribute(attr)

    def test_attr_int(self) -> None:
        # integer
        attr = helper.make_attribute("int", 3)
        self.assertEqual(attr.name, "int")
        self.assertEqual(attr.i, 3)
        checker.check_attribute(attr)
        # long integer
        attr = helper.make_attribute("int", 5)
        self.assertEqual(attr.name, "int")
        self.assertEqual(attr.i, 5)
        checker.check_attribute(attr)
        # octinteger
        attr = helper.make_attribute("int", 0o1701)
        self.assertEqual(attr.name, "int")
        self.assertEqual(attr.i, 0o1701)
        checker.check_attribute(attr)
        # hexinteger
        attr = helper.make_attribute("int", 0x1701)
        self.assertEqual(attr.name, "int")
        self.assertEqual(attr.i, 0x1701)
        checker.check_attribute(attr)

    def test_attr_doc_string(self) -> None:
        attr = helper.make_attribute("a", "value")
        self.assertEqual(attr.name, "a")
        self.assertEqual(attr.doc_string, "")
        attr = helper.make_attribute("a", "value", "doc")
        self.assertEqual(attr.name, "a")
        self.assertEqual(attr.doc_string, "doc")

    def test_attr_string(self) -> None:
        # bytes
        attr = helper.make_attribute("str", b"test")
        self.assertEqual(attr.name, "str")
        self.assertEqual(attr.s, b"test")
        checker.check_attribute(attr)
        # unspecified
        attr = helper.make_attribute("str", "test")
        self.assertEqual(attr.name, "str")
        self.assertEqual(attr.s, b"test")
        checker.check_attribute(attr)
        # unicode
        attr = helper.make_attribute("str", "test")
        self.assertEqual(attr.name, "str")
        self.assertEqual(attr.s, b"test")
        checker.check_attribute(attr)
        # empty str
        attr = helper.make_attribute("str", "")
        self.assertEqual(attr.name, "str")
        self.assertEqual(helper.get_attribute_value(attr), b"")
        checker.check_attribute(attr)

    def test_attr_repeated_float(self) -> None:
        attr = helper.make_attribute("floats", [1.0, 2.0])
        self.assertEqual(attr.name, "floats")
        self.assertEqual(list(attr.floats), [1.0, 2.0])
        checker.check_attribute(attr)

    def test_attr_repeated_int(self) -> None:
        attr = helper.make_attribute("ints", [1, 2])
        self.assertEqual(attr.name, "ints")
        self.assertEqual(list(attr.ints), [1, 2])
        checker.check_attribute(attr)

    def test_attr_repeated_mixed_floats_and_ints(self) -> None:
        attr = helper.make_attribute("mixed", [1, 2, 3.0, 4.5])
        self.assertEqual(attr.name, "mixed")
        self.assertEqual(list(attr.floats), [1.0, 2.0, 3.0, 4.5])
        checker.check_attribute(attr)

    def test_attr_repeated_str(self) -> None:
        attr = helper.make_attribute("strings", ["str1", "str2"])
        self.assertEqual(attr.name, "strings")
        self.assertEqual(list(attr.strings), [b"str1", b"str2"])
        checker.check_attribute(attr)

    def test_attr_repeated_tensor_proto(self) -> None:
        tensors = [
            helper.make_tensor(
                name='a',
                data_type=TensorProto.FLOAT,
                dims=(1,),
                vals=np.ones(1)
            ),
            helper.make_tensor(
                name='b',
                data_type=TensorProto.FLOAT,
                dims=(1,),
                vals=np.ones(1)
            )]
        attr = helper.make_attribute("tensors", tensors)
        self.assertEqual(attr.name, "tensors")
        self.assertEqual(list(attr.tensors), tensors)
        checker.check_attribute(attr)

    def test_attr_sparse_tensor_proto(self) -> None:
        dense_shape = [3, 3]
        sparse_values = [1.764052391052246, 0.40015721321105957, 0.978738009929657]
        values_tensor = helper.make_tensor(name='sparse_values', data_type=TensorProto.FLOAT,
                                        dims=[len(sparse_values)],
                                        vals=np.array(sparse_values).astype(np.float32), raw=False)

        linear_indices = [2, 3, 5]
        indices_tensor = helper.make_tensor(name='indices', data_type=TensorProto.INT64,
                                        dims=[len(linear_indices)],
                                        vals=np.array(linear_indices).astype(np.int64), raw=False)
        sparse_tensor = helper.make_sparse_tensor(values_tensor, indices_tensor, dense_shape)

        attr = helper.make_attribute("sparse_attr", sparse_tensor)
        self.assertEqual(attr.name, "sparse_attr")
        checker.check_sparse_tensor(helper.get_attribute_value(attr))
        checker.check_attribute(attr)

    def test_attr_sparse_tensor_repeated_protos(self) -> None:
        dense_shape = [3, 3]
        sparse_values = [1.764052391052246, 0.40015721321105957, 0.978738009929657]
        values_tensor = helper.make_tensor(name='sparse_values', data_type=TensorProto.FLOAT,
                                        dims=[len(sparse_values)],
                                        vals=np.array(sparse_values).astype(np.float32), raw=False)

        linear_indices = [2, 3, 5]
        indices_tensor = helper.make_tensor(name='indices', data_type=TensorProto.INT64,
                                        dims=[len(linear_indices)],
                                        vals=np.array(linear_indices).astype(np.int64), raw=False)
        sparse_tensor = helper.make_sparse_tensor(values_tensor, indices_tensor, dense_shape)

        repeated_sparse = [sparse_tensor, sparse_tensor]
        attr = helper.make_attribute("sparse_attrs", repeated_sparse)
        self.assertEqual(attr.name, "sparse_attrs")
        checker.check_attribute(attr)
        for s in helper.get_attribute_value(attr):
            checker.check_sparse_tensor(s)

    def test_attr_repeated_graph_proto(self) -> None:
        graphs = [GraphProto(), GraphProto()]
        graphs[0].name = "a"
        graphs[1].name = "b"
        attr = helper.make_attribute("graphs", graphs)
        self.assertEqual(attr.name, "graphs")
        self.assertEqual(list(attr.graphs), graphs)
        checker.check_attribute(attr)

    def test_attr_type_proto(self) -> None:
        # type_proto
        type = TypeProto()
        attr = helper.make_attribute("type_proto", type)
        self.assertEqual(attr.name, "type_proto")
        self.assertEqual(attr.tp, type)
        self.assertEqual(attr.type, AttributeProto.TYPE_PROTO)
        # type_protos
        types = [TypeProto(), TypeProto()]
        attr = helper.make_attribute("type_protos", types)

        self.assertEqual(attr.name, "type_protos")
        self.assertEqual(list(attr.type_protos), types)
        self.assertEqual(attr.type, AttributeProto.TYPE_PROTOS)

    def test_is_attr_legal(self) -> None:
        # no name, no field
        attr = AttributeProto()
        self.assertRaises(checker.ValidationError, checker.check_attribute, attr)
        # name, but no field
        attr = AttributeProto()
        attr.name = "test"
        self.assertRaises(checker.ValidationError, checker.check_attribute, attr)
        # name, with two fields
        attr = AttributeProto()
        attr.name = "test"
        attr.f = 1.0
        attr.i = 2
        self.assertRaises(checker.ValidationError, checker.check_attribute, attr)

    def test_is_attr_legal_verbose(self) -> None:

        def _set(attr: AttributeProto, type: AttributeProto.AttributeType, var: str, value: Any) -> None:
            setattr(attr, var, value)
            setattr(attr, 'type', type)

        def _extend(attr: AttributeProto, type: AttributeProto.AttributeType, var: List[Any], value: Any) -> None:
            var.extend(value)
            setattr(attr, 'type', type)

        SET_ATTR = [
            (lambda attr: _set(attr, AttributeProto.FLOAT, "f", 1.0)),
            (lambda attr: _set(attr, AttributeProto.INT, "i", 1)),
            (lambda attr: _set(attr, AttributeProto.STRING, "s", b"str")),
            (lambda attr: _extend(attr, AttributeProto.FLOATS, attr.floats, [1.0, 2.0])),
            (lambda attr: _extend(attr, AttributeProto.INTS, attr.ints, [1, 2])),
            (lambda attr: _extend(attr, AttributeProto.STRINGS, attr.strings, [b"a", b"b"])),
        ]
        # Randomly set one field, and the result should be legal.
        for _i in range(100):
            attr = AttributeProto()
            attr.name = "test"
            random.choice(SET_ATTR)(attr)
            checker.check_attribute(attr)
        # Randomly set two fields, and then ensure helper function catches it.
        for _i in range(100):
            attr = AttributeProto()
            attr.name = "test"
            for func in random.sample(SET_ATTR, 2):
                func(attr)
            self.assertRaises(checker.ValidationError,
                              checker.check_attribute,
                              attr)


class TestHelperNodeFunctions(unittest.TestCase):

    def test_node_no_arg(self) -> None:
        self.assertTrue(defs.has("Relu"))
        node_def = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        self.assertEqual(node_def.op_type, "Relu")
        self.assertEqual(node_def.name, "test")
        self.assertEqual(list(node_def.input), ["X"])
        self.assertEqual(list(node_def.output), ["Y"])

    def test_attr_doc_string(self) -> None:
        node_def = helper.make_node(
            "Relu", ["X"], ["Y"], name="test", doc_string="doc")
        self.assertEqual(node_def.doc_string, "doc")

    def test_node_with_arg(self) -> None:
        self.assertTrue(defs.has("Relu"))
        # Note: Relu actually does not need an arg, but let's
        # test it.
        node_def = helper.make_node(
            "Relu", ["X"], ["Y"],
            arg_value=1)
        self.assertEqual(node_def.op_type, "Relu")
        self.assertEqual(list(node_def.input), ["X"])
        self.assertEqual(list(node_def.output), ["Y"])
        self.assertEqual(len(node_def.attribute), 1)
        self.assertEqual(
            node_def.attribute[0],
            helper.make_attribute("arg_value", 1))

    def test_node_domain(self) -> None:
        node_def = helper.make_node(
            "Relu", ["X"], ["Y"], name="test", doc_string="doc", domain="test.domain")
        self.assertEqual(node_def.domain, "test.domain")

    def test_graph(self) -> None:
        node_def1 = helper.make_node(
            "Relu", ["X"], ["Y"])
        node_def2 = helper.make_node(
            "Add", ["X", "Y"], ["Z"])
        value_info = [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])]
        graph = helper.make_graph(
            [node_def1, node_def2],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])],
            doc_string=None,
            value_info=value_info,
        )
        self.assertEqual(graph.name, "test")
        self.assertEqual(len(graph.node), 2)
        self.assertEqual(graph.node[0], node_def1)
        self.assertEqual(graph.node[1], node_def2)
        self.assertEqual(graph.doc_string, "")
        self.assertEqual(graph.value_info[0], value_info[0])

    def test_graph_docstring(self) -> None:
        graph = helper.make_graph([], "my graph", [], [], None, "my docs")
        self.assertEqual(graph.name, "my graph")
        self.assertEqual(graph.doc_string, "my docs")

    def test_model(self) -> None:
        node_def = helper.make_node(
            "Relu", ["X"], ["Y"])
        graph_def = helper.make_graph(
            [node_def],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        self.assertRaises(AttributeError, helper.make_model, graph_def, xxx=1)
        model_def = helper.make_model(graph_def, producer_name='test')
        self.assertEqual(model_def.producer_name, 'test')

    def test_model_docstring(self) -> None:
        graph = helper.make_graph([], "my graph", [], [])
        model_def = helper.make_model(graph, doc_string='test')
        # models may have their own documentation, but don't have a name
        # their name is the domain-qualified name of the underlying graph.
        self.assertFalse(hasattr(model_def, "name"))
        self.assertEqual(model_def.doc_string, 'test')

    def test_model_metadata_props(self) -> None:
        graph = helper.make_graph([], "my graph", [], [])
        model_def = helper.make_model(graph, doc_string='test')
        helper.set_model_props(model_def, {'Title': 'my graph', 'Keywords': 'test;graph'})
        checker.check_model(model_def)
        helper.set_model_props(model_def, {'Title': 'my graph', 'Keywords': 'test;graph'})
        checker.check_model(model_def)  # helper replaces, so no dupe

        dupe = model_def.metadata_props.add()
        dupe.key = 'Title'
        dupe.value = 'Other'
        self.assertRaises(checker.ValidationError, checker.check_model, model_def)

    def test_model_irversion(self) -> None:
        def mk_model(opset_versions: List[Tuple[str, int]]) -> ModelProto:
            graph = helper.make_graph([], "my graph", [], [])
            return helper.make_model_gen_version(
                graph,
                opset_imports=[helper.make_opsetid(*pair) for pair in opset_versions])

        def test(opset_versions: List[Tuple[str, int]], ir_version: int) -> None:
            model = mk_model(opset_versions)
            self.assertEqual(model.ir_version, ir_version)
        # opset version 9 requires minimum ir_version 4
        test([("", 9)], 4)
        test([("", 10)], 5)
        test([("", 11)], 6)
        test([("", 12)], 7)
        test([("", 13)], 7)
        test([("", 14)], 7)
        test([("", 15)], 8)
        test([("", 16)], 8)
        test([("", 17)], 8)
        # standard opset can be referred to using empty-string or "ai.onnx"
        test([("ai.onnx", 9)], 4)
        test([("ai.onnx.ml", 2)], 6)
        test([("ai.onnx.ml", 3)], 8)
        test([("ai.onnx.training", 1)], 7)
        # helper should pick *max* IR version required from all opsets specified.
        test([("", 10), ("ai.onnx.ml", 2)], 6)
        self.assertRaises(ValueError, mk_model, [("", 100)])


class TestHelperTensorFunctions(unittest.TestCase):

    def test_make_tensor(self) -> None:
        np_array = np.random.randn(2, 3).astype(np.float32)

        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT,
            dims=(2, 3),
            vals=np_array
        )
        self.assertEqual(tensor.name, 'test')
        np.testing.assert_equal(np_array, numpy_helper.to_array(tensor))

        # use raw_data field to store the data
        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT,
            dims=(2, 3),
            vals=np_array.tobytes(),
            raw=True,
        )
        np.testing.assert_equal(np_array, numpy_helper.to_array(tensor))

        string_list = list(s.encode('utf-8') for s in ['Amy', 'Billy', 'Cindy', 'David'])
        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.STRING,
            dims=(2, 2),
            vals=string_list,
            raw=False
        )
        self.assertEqual(string_list, list(tensor.string_data))

    def test_make_int8_tensor(self) -> None:
        np_array = np.random.randn(2, 3).astype(np.int8)

        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.INT8,
            dims=(2, 3),
            vals=np_array
        )
        self.assertEqual(tensor.name, 'test')
        np.testing.assert_equal(np_array, numpy_helper.to_array(tensor))

        # use raw_data field to store the data
        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.INT8,
            dims=(2, 3),
            vals=np_array.tobytes(),
            raw=True,
        )
        np.testing.assert_equal(np_array, numpy_helper.to_array(tensor))

    def test_make_float16_tensor(self) -> None:
        np_array = np.random.randn(2, 3).astype(np.float16)

        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT16,
            dims=np_array.shape,
            vals=np_array
        )
        self.assertEqual(tensor.name, 'test')
        np.testing.assert_equal(np_array, numpy_helper.to_array(tensor))

    def test_make_float16_tensor_raw(self) -> None:
        np_array = np.random.randn(2, 3).astype(np.float16)

        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT16,
            dims=np_array.shape,
            vals=np_array.view(dtype=np.uint16).flatten().tobytes(),
            raw=True
        )
        self.assertEqual(tensor.name, 'test')
        np.testing.assert_equal(np_array, numpy_helper.to_array(tensor))

    def test_make_bfloat16_tensor(self) -> None:
        # numpy doesn't support bf16, so we have to compute the correct result manually
        np_array = np.array([[1.0, 2.0], [3.0, 4.0], [0.099853515625, 0.099365234375], [0.0998535081744, 0.1], [np.nan, np.inf]],
            dtype=np.float32)
        np_results = np.array([
            [struct.unpack('!f', bytes.fromhex('3F800000'))[0],   # 1.0
             struct.unpack('!f', bytes.fromhex('40000000'))[0]],  # 2.0
            [struct.unpack('!f', bytes.fromhex('40400000'))[0],   # 3.0
             struct.unpack('!f', bytes.fromhex('40800000'))[0]],  # 4.0
            [struct.unpack('!f', bytes.fromhex('3DCC0000'))[0],   # round-to-nearest-even rounds down (0x8000)
             struct.unpack('!f', bytes.fromhex('3DCC0000'))[0]],  # round-to-nearest-even rounds up   (0x8000)
            [struct.unpack('!f', bytes.fromhex('3DCC0000'))[0],   # round-to-nearest-even rounds down (0x7fff)
             struct.unpack('!f', bytes.fromhex('3DCD0000'))[0]],  # round-to-nearest-even rounds up   (0xCCCD)
            [struct.unpack('!f', bytes.fromhex('7FC00000'))[0],   # NaN
             struct.unpack('!f', bytes.fromhex('7F800000'))[0]],  # inf
        ])

        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.BFLOAT16,
            dims=np_array.shape,
            vals=np_array
        )
        self.assertEqual(tensor.name, 'test')
        np.testing.assert_equal(np_results, numpy_helper.to_array(tensor))

    def test_make_bfloat16_tensor_raw(self) -> None:
        # numpy doesn't support bf16, so we have to compute the correct result manually
        np_array = np.array([[1.0, 2.0], [3.0, 4.0], [0.099853515625, 0.099365234375], [0.0998535081744, 0.1], [np.nan, np.inf]],
            dtype=np.float32)
        np_results = np.array([
            [struct.unpack('!f', bytes.fromhex('3F800000'))[0],   # 1.0
             struct.unpack('!f', bytes.fromhex('40000000'))[0]],  # 2.0
            [struct.unpack('!f', bytes.fromhex('40400000'))[0],   # 3.0
             struct.unpack('!f', bytes.fromhex('40800000'))[0]],  # 4.0
            [struct.unpack('!f', bytes.fromhex('3DCC0000'))[0],   # truncated
             struct.unpack('!f', bytes.fromhex('3DCB0000'))[0]],  # truncated
            [struct.unpack('!f', bytes.fromhex('3DCC0000'))[0],   # truncated
             struct.unpack('!f', bytes.fromhex('3DCC0000'))[0]],  # truncated
            [struct.unpack('!f', bytes.fromhex('7FC00000'))[0],   # NaN
             struct.unpack('!f', bytes.fromhex('7F800000'))[0]],  # inf
        ])

        # write out 16-bit of fp32 to create bf16 using truncation, no rounding
        truncate = lambda x: x >> 16  # noqa: E731
        values_as_ints = np_array.astype(np.float32).view(np.uint32).flatten()
        packed_values = truncate(values_as_ints).astype(np.uint16).tobytes()
        tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.BFLOAT16,
            dims=np_array.shape,
            vals=packed_values,
            raw=True
        )
        self.assertEqual(tensor.name, 'test')
        np.testing.assert_equal(np_results, numpy_helper.to_array(tensor))

    def test_make_sparse_tensor(self) -> None:
        values = [1.1, 2.2, 3.3, 4.4, 5.5]
        values_tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT,
            dims=(5, ),
            vals=values
        )
        indices = [1, 3, 5, 7, 9]
        indices_tensor = helper.make_tensor(
            name='test_indices',
            data_type=TensorProto.INT64,
            dims=(5, ),
            vals=indices
        )
        dense_shape = [10]
        sparse = helper.make_sparse_tensor(values_tensor, indices_tensor, dense_shape)
        self.assertEqual(sparse.values, values_tensor)
        self.assertEqual(sparse.indices, indices_tensor)
        self.assertEqual(sparse.dims, dense_shape)

    def test_make_tensor_value_info(self) -> None:
        vi = helper.make_tensor_value_info('X', TensorProto.FLOAT, (2, 4))
        checker.check_value_info(vi)

        # scalar value
        vi = helper.make_tensor_value_info('Y', TensorProto.FLOAT, ())
        checker.check_value_info(vi)

    def test_make_sparse_tensor_value_info(self) -> None:
        vi = helper.make_sparse_tensor_value_info('X', TensorProto.FLOAT, (2, 3))
        checker.check_value_info(vi)

        # scalar value
        vi = helper.make_sparse_tensor_value_info('Y', TensorProto.FLOAT, ())
        checker.check_value_info(vi)


class TestHelperOptionalAndSequenceFunctions(unittest.TestCase):
    def test_make_optional(self) -> None:
        values = [1.1, 2.2, 3.3, 4.4, 5.5]
        values_tensor = helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT,
            dims=(5,),
            vals=values
        )
        optional = helper.make_optional(
            name='test',
            elem_type=OptionalProto.TENSOR,
            value=values_tensor
        )
        self.assertEqual(optional.name, 'test')
        self.assertEqual(optional.elem_type, OptionalProto.TENSOR)
        self.assertEqual(optional.tensor_value, values_tensor)

        # Test Sequence
        values_sequence = helper.make_sequence(
            name='test',
            elem_type=SequenceProto.TENSOR,
            values=[values_tensor, values_tensor]
        )
        optional = helper.make_optional(
            name='test',
            elem_type=OptionalProto.SEQUENCE,
            value=values_sequence
        )
        self.assertEqual(optional.name, 'test')
        self.assertEqual(optional.elem_type, OptionalProto.SEQUENCE)
        self.assertEqual(optional.sequence_value, values_sequence)

        # Test None
        optional_none = helper.make_optional(
            name='test',
            elem_type=OptionalProto.UNDEFINED,
            value=None
        )
        self.assertEqual(optional_none.name, 'test')
        self.assertEqual(optional_none.elem_type, OptionalProto.UNDEFINED)
        self.assertFalse(optional_none.HasField('tensor_value'))

    def test_make_optional_value_info(self) -> None:
        tensor_type_proto = helper.make_tensor_type_proto(elem_type=2, shape=[5])
        tensor_val_into = helper.make_value_info(
            name='test',
            type_proto=tensor_type_proto)
        optional_type_proto = helper.make_optional_type_proto(tensor_type_proto)
        optional_val_info = helper.make_value_info(
            name='test',
            type_proto=optional_type_proto)

        self.assertEqual(optional_val_info.name, 'test')
        self.assertTrue(optional_val_info.type.optional_type)
        self.assertEqual(optional_val_info.type.optional_type.elem_type, tensor_val_into.type)

        # Test Sequence
        sequence_type_proto = helper.make_sequence_type_proto(tensor_type_proto)
        optional_type_proto = helper.make_optional_type_proto(sequence_type_proto)
        optional_val_info = helper.make_value_info(
            name='test',
            type_proto=optional_type_proto)

        self.assertEqual(optional_val_info.name, 'test')
        self.assertTrue(optional_val_info.type.optional_type)
        sequence_value_info = helper.make_value_info(
            name='test',
            type_proto=tensor_type_proto)
        self.assertEqual(optional_val_info.type.optional_type.elem_type.sequence_type.elem_type, sequence_value_info.type)

    def test_make_seuence_value_info(self) -> None:
        tensor_type_proto = helper.make_tensor_type_proto(elem_type=2, shape=None)
        sequence_type_proto = helper.make_sequence_type_proto(tensor_type_proto)
        sequence_val_info = helper.make_value_info(
            name='test',
            type_proto=sequence_type_proto
        )
        sequence_val_info_prim = helper.make_tensor_sequence_value_info(
            name='test',
            elem_type=2,
            shape=None)

        self.assertEqual(sequence_val_info, sequence_val_info_prim)


class TestPrintableGraph(unittest.TestCase):

    def test_initializer_with_matching_graph_input(self) -> None:
        add = helper.make_node("Add", ["X", "Y_Initializer"], ["Z"])
        value_info = [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1])]

        graph = helper.make_graph(
            [add],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1]),
             helper.make_tensor_value_info("Y_Initializer", TensorProto.FLOAT, [1])],  # inputs
            [helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1])],  # outputs
            [helper.make_tensor("Y_Initializer", TensorProto.FLOAT, [1], [1])],  # initializers
            doc_string=None,
            value_info=value_info
        )

        graph_str = helper.printable_graph(graph)
        self.assertTrue(''') optional inputs with matching initializers (
  %Y_Initializer[FLOAT, 1]''' in graph_str, graph_str)

    def test_initializer_no_matching_graph_input(self) -> None:
        add = helper.make_node("Add", ["X", "Y_Initializer"], ["Z"])
        value_info = [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1])]

        graph = helper.make_graph(
            [add],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1])],  # inputs
            [helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1])],  # outputs
            [helper.make_tensor("Y_Initializer", TensorProto.FLOAT, [1], [1])],  # initializers
            doc_string=None,
            value_info=value_info
        )

        graph_str = helper.printable_graph(graph)
        self.assertTrue(''') initializers (
  %Y_Initializer[FLOAT, 1]''' in graph_str, graph_str)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

# ATTENTION: The code in this file is highly EXPERIMENTAL.
# Adventurous users should note that the APIs will probably change.

"""onnx optimizer

This enables users to optimize their models.
"""

import onnx
import onnxoptimizer.onnx_opt_cpp2py_export as C
from .version import version as __version__  # noqa
from onnx import ModelProto
from typing import Text, Sequence, Optional

import tempfile
import os

get_available_passes = C.get_available_passes

get_fuse_and_elimination_passes = C.get_fuse_and_elimination_passes


def optimize(model, passes=None, fixed_point=False):  # type: (ModelProto, Optional[Sequence[Text]], bool) -> ModelProto
    """Apply the optimization on the serialized ModelProto.

    Arguments:
        model (ModelProto): model
        passes (list of string): list of optimization names

    Return:
        return (ModelProto) optimized model
    """

    if passes is None:
        print('WARNING: defualt optimization passes will be enlarged to all fuse and elimination passes in the next version')
        passes = ['eliminate_nop_transpose',
                  'eliminate_nop_pad',
                  'fuse_consecutive_transposes',
                  'fuse_transpose_into_gemm']
    if not isinstance(model, ModelProto):
        raise ValueError(
            'Optimizer only accepts ModelProto, incorrect type: {}'.format(type(model)))
    try:
        model_str = model.SerializeToString()
        if fixed_point:
            optimized_model_str = C.optimize_fixedpoint(model_str, passes)
        else:
            optimized_model_str = C.optimize(model_str, passes)

        return onnx.load_from_string(optimized_model_str)
    except ValueError:
        file_src = tempfile.NamedTemporaryFile(suffix=".onnx", delete=False)
        file_dest = tempfile.NamedTemporaryFile(suffix=".onnx", delete=False)
        data_file_src = tempfile.NamedTemporaryFile(delete=False)
        data_file_dest = tempfile.NamedTemporaryFile(delete=False)
        data_src_rel_filename = os.path.relpath(data_file_src.name, os.path.dirname(file_src.name))
        data_dest_rel_filename = os.path.relpath(data_file_dest.name, os.path.dirname(file_dest.name))
        try:
            onnx.save(model, file_src.name, save_as_external_data=True, location=data_src_rel_filename, convert_attribute=True,)
            if fixed_point:
                C.optimize_fixedpoint_from_path(file_src.name, file_dest.name, passes, data_dest_rel_filename)
            else:
                C.optimize_from_path(file_src.name, file_dest.name, passes, data_dest_rel_filename)
            return onnx.load(file_dest, load_external_data=True)
        finally:
            os.remove(file_src.name)
            os.remove(file_dest.name)
            os.remove(data_file_src.name)
            os.remove(data_file_dest.name)


__all__ = ['optimize', 'get_available_passes', 'get_fuse_and_elimination_passes']



#!/usr/bin/env python
# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
# https://developers.google.com/protocol-buffers/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A conformance test implementation for the Python protobuf library.

See conformance.proto for more information.
"""

import struct
import sys
import os
from google.protobuf import json_format
from google.protobuf import message
from google.protobuf import test_messages_proto3_pb2
from google.protobuf import test_messages_proto2_pb2
from google.protobuf import text_format
import conformance_pb2

sys.stdout = os.fdopen(sys.stdout.fileno(), 'wb', 0)
sys.stdin = os.fdopen(sys.stdin.fileno(), 'rb', 0)

test_count = 0
verbose = False

class ProtocolError(Exception):
  pass

def do_test(request):
  response = conformance_pb2.ConformanceResponse()

  if request.message_type == "conformance.FailureSet":
    failure_set = conformance_pb2.FailureSet()
    failures = []
    # TODO(gerbens): Remove, this is a hack to detect if the old vs new
    # parser is used by the cpp code. Relying on a bug in the old parser.
    hack_proto = test_messages_proto2_pb2.TestAllTypesProto2()
    old_parser = True
    try:
      hack_proto.ParseFromString(b"\322\002\001")
    except message.DecodeError as e:
      old_parser = False
    if old_parser:
      # the string above is one of the failing conformance test strings of the
      # old parser. If we succeed the c++ implementation is using the old
      # parser so we add the list of failing conformance tests.
      failures = [
          "Required.Proto3.ProtobufInput.PrematureEofInDelimitedDataForKnownNonRepeatedValue.MESSAGE",
          "Required.Proto3.ProtobufInput.PrematureEofInDelimitedDataForKnownRepeatedValue.MESSAGE",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.BOOL",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.DOUBLE",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.ENUM",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.FIXED32",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.FIXED64",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.FLOAT",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.INT32",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.INT64",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.SFIXED32",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.SFIXED64",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.SINT32",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.SINT64",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.UINT32",
          "Required.Proto3.ProtobufInput.PrematureEofInPackedField.UINT64",
          "Required.Proto2.ProtobufInput.PrematureEofInDelimitedDataForKnownNonRepeatedValue.MESSAGE",
          "Required.Proto2.ProtobufInput.PrematureEofInDelimitedDataForKnownRepeatedValue.MESSAGE",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.BOOL",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.DOUBLE",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.ENUM",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.FIXED32",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.FIXED64",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.FLOAT",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.INT32",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.INT64",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.SFIXED32",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.SFIXED64",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.SINT32",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.SINT64",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.UINT32",
          "Required.Proto2.ProtobufInput.PrematureEofInPackedField.UINT64",
      ]
    for x in failures:
      failure_set.failure.append(x)
    response.protobuf_payload = failure_set.SerializeToString()
    return response

  isProto3 = (request.message_type == "protobuf_test_messages.proto3.TestAllTypesProto3")
  isJson = (request.WhichOneof('payload') == 'json_payload')
  isProto2 = (request.message_type == "protobuf_test_messages.proto2.TestAllTypesProto2")

  if (not isProto3) and (not isJson) and (not isProto2):
    raise ProtocolError("Protobuf request doesn't have specific payload type")

  test_message = test_messages_proto2_pb2.TestAllTypesProto2() if isProto2 else \
    test_messages_proto3_pb2.TestAllTypesProto3()

  try:
    if request.WhichOneof('payload') == 'protobuf_payload':
      try:
        test_message.ParseFromString(request.protobuf_payload)
      except message.DecodeError as e:
        response.parse_error = str(e)
        return response

    elif request.WhichOneof('payload') == 'json_payload':
      try:
        ignore_unknown_fields = \
            request.test_category == \
                conformance_pb2.JSON_IGNORE_UNKNOWN_PARSING_TEST
        json_format.Parse(request.json_payload, test_message,
                          ignore_unknown_fields)
      except Exception as e:
        response.parse_error = str(e)
        return response

    elif request.WhichOneof('payload') == 'text_payload':
      try:
        text_format.Parse(request.text_payload, test_message)
      except Exception as e:
        response.parse_error = str(e)
        return response

    else:
      raise ProtocolError("Request didn't have payload.")

    if request.requested_output_format == conformance_pb2.UNSPECIFIED:
      raise ProtocolError("Unspecified output format")

    elif request.requested_output_format == conformance_pb2.PROTOBUF:
      response.protobuf_payload = test_message.SerializeToString()

    elif request.requested_output_format == conformance_pb2.JSON:
      try:
        response.json_payload = json_format.MessageToJson(test_message)
      except Exception as e:
        response.serialize_error = str(e)
        return response

    elif request.requested_output_format == conformance_pb2.TEXT_FORMAT:
      response.text_payload = text_format.MessageToString(
          test_message, print_unknown_fields=request.print_unknown_fields)

  except Exception as e:
    response.runtime_error = str(e)

  return response

def do_test_io():
  length_bytes = sys.stdin.read(4)
  if len(length_bytes) == 0:
    return False   # EOF
  elif len(length_bytes) != 4:
    raise IOError("I/O error")

  # "I" is "unsigned int", so this depends on running on a platform with
  # 32-bit "unsigned int" type.  The Python struct module unfortunately
  # has no format specifier for uint32_t.
  length = struct.unpack("<I", length_bytes)[0]
  serialized_request = sys.stdin.read(length)
  if len(serialized_request) != length:
    raise IOError("I/O error")

  request = conformance_pb2.ConformanceRequest()
  request.ParseFromString(serialized_request)

  response = do_test(request)

  serialized_response = response.SerializeToString()
  sys.stdout.write(struct.pack("<I", len(serialized_response)))
  sys.stdout.write(serialized_response)
  sys.stdout.flush()

  if verbose:
    sys.stderr.write("conformance_python: request=%s, response=%s\n" % (
                       request.ShortDebugString().c_str(),
                       response.ShortDebugString().c_str()))

  global test_count
  test_count += 1

  return True

while True:
  if not do_test_io():
    sys.stderr.write("conformance_python: received EOF from test runner " +
                     "after %s tests, exiting\n" % (test_count))
    sys.exit(0)



def function_to_text(serialized_function_proto : bytes) -> str:
    ...


def graph_to_text(serialized_graph_proto : bytes) -> str:
    ...


def model_to_text(serialized_model_proto : bytes) -> str:
    ...



# SPDX-License-Identifier: Apache-2.0
import unittest

from typing import Sequence
import numpy as np  # type: ignore

from onnx import checker, helper, numpy_helper, shape_inference
from onnx import TensorProto, GraphProto, SparseTensorProto
import onnx.onnx_cpp2py_export.checker as C
import onnx.defs


class TestChecker(unittest.TestCase):
    @property
    def _sample_float_tensor(self) -> TensorProto:
        np_array = np.random.randn(2, 3).astype(np.float32)
        return helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT,
            dims=(2, 3),
            vals=np_array.reshape(6).tolist()
        )

    def make_sparse(self,
                    shape: Sequence[int],
                    values: Sequence[int],
                    indices_shape: Sequence[int],
                    indices: Sequence[int],
                    name: str = 'spval'
                    ) -> SparseTensorProto:
        sparse = SparseTensorProto()
        sparse.dims.extend(shape)
        nnz = len(values)

        sparse.values.CopyFrom(helper.make_tensor(name, TensorProto.INT64, (nnz,), values))
        sparse.indices.CopyFrom(helper.make_tensor('spind', TensorProto.INT64, indices_shape, indices))
        return sparse

    def test_check_node(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")

        checker.check_node(node)

    def test_check_node_input_marked_optional(self) -> None:
        # GivenTensorFill's input is marked optional, hence it is used in this test.
        node = helper.make_node(
            "GivenTensorFill", [], ["Y"], name="test")
        checker.check_node(node)

        # Explicitly pass the empty string as optional
        node = helper.make_node(
            "GivenTensorFill", [""], ["Y"], name="test")

        # Input of RELU is not optional
        node = helper.make_node(
            "Relu", [""], ["Y"], name="test")
        self.assertRaises(checker.ValidationError, checker.check_node, node)

    def test_check_graph_ir_version_3(self) -> None:
        ctx = C.CheckerContext()
        ctx.ir_version = 3
        ctx.opset_imports = {'': onnx.defs.onnx_opset_version()}

        def check_ir_version_3(g: GraphProto) -> None:
            checker.check_graph(g, ctx)

        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        check_ir_version_3(graph)

        graph.initializer.extend([self._sample_float_tensor])

        graph.initializer[0].name = 'no-exist'

        self.assertRaises(checker.ValidationError, check_ir_version_3, graph)

        graph.initializer[0].name = 'X'
        check_ir_version_3(graph)

    def test_check_graph(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        checker.check_graph(graph)

        graph.initializer.extend([self._sample_float_tensor])

        graph.initializer[0].name = 'no-exist'
        checker.check_graph(graph)

        graph.initializer[0].name = 'X'
        checker.check_graph(graph)

    def test_check_graph_types(self) -> None:
        # This is for https://github.com/onnx/onnx/issues/3849.
        # It confirms that type checking is performed
        # when checker.check_model is called with full_check=True

        node_div = helper.make_node(
            "Div", ["X", "Y"], ["Z"], name="test_div")
        node_identity = helper.make_node(
            "Identity", ["Z"], ["W"], name="test_identity")

        graph = helper.make_graph(
            [node_div, node_identity],
            "test",
            [
                helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2]),
                # intentionally use a BOOL type which is not supported by the Div op.
                helper.make_tensor_value_info("Y", TensorProto.BOOL, [1, 2])],
            [helper.make_tensor_value_info("W", TensorProto.FLOAT, [1, 2])])

        model = helper.make_model(graph, producer_name='test')

        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)

        checker.check_graph(graph)

        graph = helper.make_graph(
            [node_div, node_identity],
            "test",
            [
                helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2]),
                # intentionally use a Int32 type which is in conflict with Div's other input X.
                helper.make_tensor_value_info("Y", TensorProto.INT32, [1, 2])],
            [helper.make_tensor_value_info("W", TensorProto.FLOAT, [1, 2])])

        model = helper.make_model(graph, producer_name='test')

        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)

        checker.check_graph(graph)

    def test_check_graph_empty_initializer_name(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        checker.check_graph(graph)

        # Supply no name for the initializer
        graph.initializer.extend([self._sample_float_tensor])
        graph.initializer[0].name = ''
        self.assertRaises(checker.ValidationError, checker.check_graph, graph)

    def test_check_graph_empty_sparse_initializer_name(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        checker.check_graph(graph)

        # Supply no name for the sparse_initializer
        sparse = self.make_sparse([100], [13, 17, 19], [3], [9, 27, 81], '')
        graph.sparse_initializer.extend([sparse])
        self.assertRaises(checker.ValidationError, checker.check_graph, graph)

    def test_check_graph_duplicate_init_names(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        checker.check_graph(graph)

        graph.initializer.extend([self._sample_float_tensor])
        graph.initializer[0].name = 'X'

        # Add sparse initializer with the same name as above
        sparse = self.make_sparse([100], [13, 17, 19], [3], [9, 27, 81], 'X')
        graph.sparse_initializer.extend([sparse])
        self.assertRaises(checker.ValidationError, checker.check_graph, graph)

    def test_check_graph_optional_input(self) -> None:
        # GivenTensorFill's input is marked optional, hence it is used in this test.
        node = helper.make_node(
            "GivenTensorFill", [""], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        checker.check_graph(graph)

    def test_check_graph_ssa(self) -> None:
        relu1 = helper.make_node(
            "Relu", ["X"], ["Z"], name="relu1")
        relu2 = helper.make_node(
            "Relu", ["Y"], ["Z"], name="relu2")

        graph = helper.make_graph(
            [relu1, relu2],
            "test",
            inputs=[
                helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2]),
                helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])
            ],
            outputs=[
                helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])
            ]
        )
        self.assertRaises(checker.ValidationError, checker.check_graph, graph)

    def test_check_graph_topologically_sorted(self) -> None:
        n1 = helper.make_node(
            "Scale", ["X"], ["Y"], scale=2., name="n1")
        n2 = helper.make_node(
            "Scale", ["Y"], ["Z"], scale=3., name="n2")

        graph = helper.make_graph(
            [n2, n1],
            "test",
            inputs=[
                helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])
            ],
            outputs=[
                helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])
            ]
        )
        self.assertRaises(checker.ValidationError, checker.check_graph, graph)

    def test_check_model(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        model = helper.make_model(graph, producer_name='test')

        checker.check_model(model)

    def test_check_serialized_model(self) -> None:
        node = helper.make_node(
            "Relu", ["X"], ["Y"], name="test")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        model = helper.make_model(graph, producer_name='test')

        checker.check_model(model.SerializeToString())

    def test_check_old_model(self) -> None:
        node = helper.make_node(
            "Pad", ["X"], ["Y"], paddings=(0, 0, 0, 0))
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        onnx_id = helper.make_opsetid("", 1)
        model = helper.make_model(graph, producer_name='test', opset_imports=[onnx_id])

        checker.check_model(model)

    def test_check_tensor(self) -> None:
        tensor = self._sample_float_tensor
        checker.check_tensor(tensor)

        tensor.raw_data = np.random.randn(2, 3).astype(np.float32).tobytes()
        self.assertRaises(checker.ValidationError, checker.check_tensor, tensor)

    def test_check_string_tensor(self) -> None:
        tensor = TensorProto()
        tensor.data_type = TensorProto.STRING
        tensor.dims.append(1)
        tensor.string_data.append(b'Test')
        checker.check_tensor(tensor)

        del tensor.string_data[:]
        tensor.raw_data = b'Test'
        # string data should not be stored in raw_data field
        self.assertRaises(checker.ValidationError, checker.check_tensor, tensor)

    def test_check_tensor_mismatched_field(self) -> None:
        tensor = self._sample_float_tensor
        tensor.data_type = TensorProto.INT32
        self.assertRaises(checker.ValidationError, checker.check_tensor, tensor)

    def test_nested_graph(self) -> None:
        n1 = helper.make_node(
            "Scale", ["X"], ["Y"], scale=2., name="n1")
        n2 = helper.make_node(
            "Scale", ["Y"], ["Z"], scale=3., name="n2")

        graph = helper.make_graph(
            [n1, n2],
            "nested",
            inputs=[],
            outputs=[
                helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])
            ]
        )

        i1 = helper.make_node(
            "If", ["cond"], ["Z"], then_branch=graph, else_branch=graph)

        graph = helper.make_graph(
            [i1],
            "test",
            inputs=[
                helper.make_tensor_value_info("cond", TensorProto.BOOL, [1]),
                helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])
            ],
            outputs=[helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])],
        )

        checker.check_graph(graph)

    def test_nested_graph_without_subgraph_input_shape(self) -> None:
        n1 = helper.make_node(
            "Scale", ["X"], ["Y"], scale=2., name="n1")
        n2 = helper.make_node(
            "Scale", ["Y"], ["Z"], scale=3., name="n2")

        input_x = onnx.ValueInfoProto()
        input_x.name = "X"
        graph = helper.make_graph(
            [n1, n2],
            "nested",
            inputs=[],
            outputs=[
                helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])
            ]
        )

        i1 = helper.make_node(
            "If", ["cond"], ["Z"], then_branch=graph, else_branch=graph)

        graph = helper.make_graph(
            [i1],
            "test",
            inputs=[
                helper.make_tensor_value_info("cond", TensorProto.BOOL, [1]),
                helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])
            ],
            outputs=[helper.make_tensor_value_info("Z", TensorProto.FLOAT, [1, 2])],
        )

        checker.check_graph(graph)

    @property
    def _sample_0_elem_tensor(self) -> TensorProto:
        np_array = np.random.randn(0, 3).astype(np.float32)
        return helper.make_tensor(
            name='test',
            data_type=TensorProto.FLOAT,
            dims=(0, 3),
            vals=np_array.reshape(0).tolist()
        )

    def test_check_tensor_zero_elem(self) -> None:
        tensor = self._sample_0_elem_tensor
        checker.check_tensor(tensor)

    def test_check_removed_experimental_op(self) -> None:
        node = helper.make_node(
            "ConstantFill", [], ["Y"], name="test", shape=[1, 2])
        checker.check_node(node)

    def test_skip_schema_check_on_non_standard_domain(self) -> None:
        node = helper.make_node(
            "NonExistOp", ["X"], ["Y"], name="test", domain="test.domain")
        graph = helper.make_graph(
            [node],
            "test",
            [helper.make_tensor_value_info("X", TensorProto.FLOAT, [1, 2])],
            [helper.make_tensor_value_info("Y", TensorProto.FLOAT, [1, 2])])
        onnx_id = helper.make_opsetid("test.domain", 1)
        model = helper.make_model(graph, producer_name='test',
                                  opset_imports=[onnx_id])
        checker.check_model(model)

    def test_check_sparse_tensor(self) -> None:
        sparse = self.make_sparse([100], [13, 17, 19], [3], [9, 27, 81])
        checker.check_sparse_tensor(sparse)

    def test_check_sparse_tensor_invalid_index(self) -> None:
        # index value 181 is out-of-range
        sparse = self.make_sparse([100], [13, 17, 19], [3], [9, 27, 181])
        self.assertRaises(checker.ValidationError, checker.check_sparse_tensor, sparse)

    def test_check_sparse_tensor_unordered(self) -> None:
        # index values are not in sorted order
        sparse = self.make_sparse([100], [13, 17, 19], [3], [27, 9, 81])
        self.assertRaises(checker.ValidationError, checker.check_sparse_tensor, sparse)

    def test_check_sparse_tensor_coo_format(self) -> None:
        sparse = self.make_sparse([10, 10], [13, 17, 19], [3, 2], [0, 9, 2, 7, 8, 1])
        checker.check_sparse_tensor(sparse)

    def test_check_sparse_tensor_coo_format_invalid_index(self) -> None:
        sparse = self.make_sparse([10, 10], [13, 17, 19], [3, 2], [0, 9, 0, 27, 8, 1])
        self.assertRaises(checker.ValidationError, checker.check_sparse_tensor, sparse)

    def test_check_sparse_tensor_coo_format_invalid_shape(self) -> None:
        sparse = self.make_sparse([10, 10], [13, 17, 19], [2, 3], [0, 9, 2, 7, 8, 1])
        self.assertRaises(checker.ValidationError, checker.check_sparse_tensor, sparse)

    def test_check_sparse_tensor_coo_format_invalid_dim2(self) -> None:
        sparse = self.make_sparse([10, 10], [13, 17, 19], [3, 1], [0, 1, 2])
        self.assertRaises(checker.ValidationError, checker.check_sparse_tensor, sparse)

    def test_check_sparse_matmul(self) -> None:
        M = 5
        N = 10
        # Create ValueInfoProto for input X of shape [N]
        X = helper.make_tensor_value_info('X', TensorProto.FLOAT, [N])
        # Create a [M,N] sparse-matrix constant C
        sparse_tensor = self.make_sparse([M, N], [2, 3, 1], [3], [3, 11, 37])
        node1 = helper.make_node('Constant', [], ['C'], sparse_value=sparse_tensor)
        # Create ValueInfoProto for output Y of shape [M]
        Y = helper.make_tensor_value_info('Y', TensorProto.FLOAT, [M])
        # Compute Y = C X
        node2 = helper.make_node('MatMul', ['C', 'X'], ['Y'])
        # create graph
        graph = helper.make_graph([node1, node2], "sparse_matmul", [X], [Y])
        # check graph
        checker.check_graph(graph)

    def test_check_model_unsupported_input_type(self) -> None:
        N = 10
        X = helper.make_tensor_value_info('X', TensorProto.BOOL, [N])
        Y = helper.make_tensor_value_info('Y', TensorProto.FLOAT, [N])
        Z = helper.make_tensor_value_info('Z', TensorProto.FLOAT, [N])
        onnx_id = helper.make_opsetid("", 6)
        node = helper.make_node('Add', ['X', 'Y'], ['Z'])
        graph = helper.make_graph([node], "test_add_input", [X, Y], [Z])
        model = helper.make_model(graph, producer_name='test', opset_imports=[onnx_id])
        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)

    def test_check_model_inconsistent_type(self) -> None:
        N = 10
        X = helper.make_tensor_value_info('X', TensorProto.FLOAT, [N])
        Y = helper.make_tensor_value_info('Y', TensorProto.INT32, [N])
        Z = helper.make_tensor_value_info('Z', TensorProto.FLOAT, [N])
        onnx_id = helper.make_opsetid("", 6)
        node = helper.make_node('Add', ['X', 'Y'], ['Z'])
        graph = helper.make_graph([node], "test_add_input", [X, Y], [Z])
        model = helper.make_model(graph, producer_name='test', opset_imports=[onnx_id])
        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)

    def test_check_model_unsupported_output_type(self) -> None:
        N = 10
        X = helper.make_tensor_value_info('X', TensorProto.FLOAT, [N])
        Y = helper.make_tensor_value_info('Y', TensorProto.FLOAT, [N])
        Z = helper.make_tensor_value_info('Z', TensorProto.BOOL, [N])
        onnx_id = helper.make_opsetid("", 6)
        node = helper.make_node('Add', ['X', 'Y'], ['Z'])
        graph = helper.make_graph([node], "test_add_input", [X, Y], [Z])
        model = helper.make_model(graph, producer_name='test', opset_imports=[onnx_id])
        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)

    def test_loop_with_same_initializer_input_below_ir4(self) -> None:
        # This is for testing IR<4: tensors must exist both in initializer and input
        # shape_inference should allow different number of graph input and node input for Loop
        # Comes from a tf2onnx model

        model = helper.make_model(
            opset_imports=[helper.make_operatorsetid('', 8)],
            ir_version=3,
            graph=helper.make_graph(
                name='test-loop',
                inputs=[
                    helper.make_tensor_value_info('input_0', TensorProto.INT32, shape=[1]),
                    helper.make_tensor_value_info('while_maximum_iterations_0', TensorProto.INT64, shape=[]),
                    helper.make_tensor_value_info('const_fold_opt__18', TensorProto.INT64, shape=[1]),
                    helper.make_tensor_value_info('const_fold_opt__17', TensorProto.FLOAT, shape=[]),
                    helper.make_tensor_value_info('Const_0', TensorProto.INT32, shape=[1]),
                ],
                outputs=[helper.make_tensor_value_info('output_0', TensorProto.INT32, shape=[1])],
                initializer=[
                    numpy_helper.from_array(np.array(9223372036854775807, dtype=np.int64), name='while_maximum_iterations_0'),
                    numpy_helper.from_array(np.array([-1], dtype=np.int64), name='const_fold_opt__18'),
                    numpy_helper.from_array(np.array(10.0, dtype=np.float32), name='const_fold_opt__17'),
                    numpy_helper.from_array(np.array([1], dtype=np.int32), name='Const_0'),
                ],
                nodes=[
                    helper.make_node('Cast', inputs=['input_0'], outputs=['while_cond_158_while_Less__13_0'], name='while_cond_158_while_Less__13', domain='', to=TensorProto.FLOAT),
                    helper.make_node('Less', inputs=['while_cond_158_while_Less__13_0', 'const_fold_opt__17'], outputs=['while_cond_158_while_Less_0'], name='while_cond_158_while_Less', domain=''),
                    helper.make_node('Squeeze', inputs=['while_cond_158_while_Less_0'], outputs=['while_cond_158_while_Squeeze_0'], name='while_cond_158_while_Squeeze', domain=''),
                    helper.make_node(
                        'Loop',
                        inputs=['while_maximum_iterations_0', 'while_cond_158_while_Squeeze_0', 'input_0', 'Const_0'],
                        outputs=['while_loop_0', 'while_loop_1'],
                        name='while_loop',
                        body=helper.make_graph(
                            name='while_body',
                            inputs=[
                                helper.make_tensor_value_info('while_while_loop_counter_0', TensorProto.INT64, shape=[]),
                                helper.make_tensor_value_info('cond__15_0', TensorProto.BOOL, shape=[]),
                                helper.make_tensor_value_info('while_placeholder_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('while_add_const_0_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('const_fold_opt__19', TensorProto.FLOAT, shape=[]),
                            ],
                            outputs=[
                                helper.make_tensor_value_info('cond___while_Identity_graph_outputs_Identity__3_0', TensorProto.BOOL, shape=[]),
                                helper.make_tensor_value_info('while_Identity_2_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('while_add_const_0_0', TensorProto.INT32, shape=[1]),
                            ],
                            initializer=[numpy_helper.from_array(np.array(10.0, dtype=np.float32), name='const_fold_opt__19')],
                            nodes=[
                                helper.make_node('Add', inputs=['while_placeholder_0', 'while_add_const_0_0'], outputs=['while_Identity_2_0'], name='while_Add'),
                                helper.make_node('Cast', inputs=['while_Identity_2_0'], outputs=['cond___while_Less__13_0'], name='cond___while_Less__13', domain='', to=TensorProto.FLOAT),
                                helper.make_node('Less', inputs=['cond___while_Less__13_0', 'const_fold_opt__19'], outputs=['cond___while_Less_0'], name='cond___while_Less', domain=''),
                                helper.make_node('Squeeze', inputs=['cond___while_Less_0'], outputs=['cond___while_Identity_graph_outputs_Identity__3_0'], name='cond___while_Squeeze', domain=''),
                            ],
                        ),
                    ),
                    helper.make_node('Unsqueeze', inputs=['while_loop_0'], outputs=['Reshape_tensor_0'], name='Reshape_tensor', axes=[0]),
                    helper.make_node('Reshape', inputs=['Reshape_tensor_0', 'const_fold_opt__18'], outputs=['output_0'], name='Reshape'),
                ],
            ),
        )
        # Should not throw an error
        checker.check_model(model, full_check=True)

    def test_loop_with_different_initializer_input_below_ir4(self) -> None:
        # This is for testing IR<4: tensors must exist both in initializer and input
        # Testing an optional input which does not exist in initializers
        # Checker should throw an error said the missing input is not in initializers

        model = helper.make_model(
            opset_imports=[helper.make_operatorsetid('', 8)],
            ir_version=3,
            graph=helper.make_graph(
                name='test-loop',
                inputs=[
                    helper.make_tensor_value_info('input_0', TensorProto.INT32, shape=[1]),
                    helper.make_tensor_value_info('while_maximum_iterations_0', TensorProto.INT64, shape=[]),
                    helper.make_tensor_value_info('const_fold_opt__18', TensorProto.INT64, shape=[1]),
                    helper.make_tensor_value_info('const_fold_opt__17', TensorProto.FLOAT, shape=[]),
                    helper.make_tensor_value_info('Const_0', TensorProto.INT32, shape=[1]),
                ],
                outputs=[helper.make_tensor_value_info('output_0', TensorProto.INT32, shape=[1])],
                initializer=[
                    numpy_helper.from_array(np.array(9223372036854775807, dtype=np.int64), name='while_maximum_iterations_0'),
                    numpy_helper.from_array(np.array([-1], dtype=np.int64), name='const_fold_opt__18'),
                    numpy_helper.from_array(np.array(10.0, dtype=np.float32), name='const_fold_opt__17'),
                    numpy_helper.from_array(np.array([1], dtype=np.int32), name='Const_0'),
                ],
                nodes=[
                    helper.make_node('Cast', inputs=['input_0'], outputs=['while_cond_158_while_Less__13_0'], name='while_cond_158_while_Less__13', domain='', to=TensorProto.FLOAT),
                    helper.make_node('Less', inputs=['while_cond_158_while_Less__13_0', 'const_fold_opt__17'], outputs=['while_cond_158_while_Less_0'], name='while_cond_158_while_Less', domain=''),
                    helper.make_node('Squeeze', inputs=['while_cond_158_while_Less_0'], outputs=['while_cond_158_while_Squeeze_0'], name='while_cond_158_while_Squeeze', domain=''),
                    helper.make_node(
                        'Loop',
                        inputs=['while_maximum_iterations_0', 'while_cond_158_while_Squeeze_0', 'input_0', 'Const_0'],
                        outputs=['while_loop_0', 'while_loop_1'],
                        name='while_loop',
                        body=helper.make_graph(
                            name='while_body',
                            inputs=[
                                helper.make_tensor_value_info('while_while_loop_counter_0', TensorProto.INT64, shape=[]),
                                helper.make_tensor_value_info('cond__15_0', TensorProto.BOOL, shape=[]),
                                helper.make_tensor_value_info('while_placeholder_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('while_add_const_0_0', TensorProto.INT32, shape=[1]),
                                # The following input cannot be found in initializer and checker should throw an error
                                helper.make_tensor_value_info('const_fold_opt__18', TensorProto.FLOAT, shape=[]),
                            ],
                            outputs=[
                                helper.make_tensor_value_info('cond___while_Identity_graph_outputs_Identity__3_0', TensorProto.BOOL, shape=[]),
                                helper.make_tensor_value_info('while_Identity_2_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('while_add_const_0_0', TensorProto.INT32, shape=[1]),
                            ],
                            initializer=[],
                            nodes=[
                                helper.make_node('Add', inputs=['while_placeholder_0', 'while_add_const_0_0'], outputs=['while_Identity_2_0'], name='while_Add'),
                                helper.make_node('Cast', inputs=['while_Identity_2_0'], outputs=['cond___while_Less__13_0'], name='cond___while_Less__13', domain='', to=TensorProto.FLOAT)
                            ],
                        ),
                    ),
                    helper.make_node('Unsqueeze', inputs=['while_loop_0'], outputs=['Reshape_tensor_0'], name='Reshape_tensor', axes=[0]),
                    helper.make_node('Reshape', inputs=['Reshape_tensor_0', 'const_fold_opt__18'], outputs=['output_0'], name='Reshape'),
                ],
            ),
        )
        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)

    def test_loop_with_same_initializer_input_above_ir4(self) -> None:
        # This is for testing IR>=4:
        # Cannot use the same name as both a subgraph initializer and subgraph input

        model = helper.make_model(
            opset_imports=[helper.make_operatorsetid('', 11)],
            ir_version=6,
            graph=helper.make_graph(
                name='test-loop',
                inputs=[
                    helper.make_tensor_value_info('input_0', TensorProto.INT32, shape=[1]),
                    helper.make_tensor_value_info('while_maximum_iterations_0', TensorProto.INT64, shape=[]),
                    helper.make_tensor_value_info('const_fold_opt__18', TensorProto.INT64, shape=[1]),
                    helper.make_tensor_value_info('const_fold_opt__17', TensorProto.FLOAT, shape=[]),
                    helper.make_tensor_value_info('Const_0', TensorProto.INT32, shape=[1]),
                ],
                outputs=[helper.make_tensor_value_info('output_0', TensorProto.INT32, shape=[1])],
                initializer=[
                    numpy_helper.from_array(np.array(9223372036854775807, dtype=np.int64), name='while_maximum_iterations_0'),
                    numpy_helper.from_array(np.array([-1], dtype=np.int64), name='const_fold_opt__18'),
                    numpy_helper.from_array(np.array(10.0, dtype=np.float32), name='const_fold_opt__17'),
                    numpy_helper.from_array(np.array([1], dtype=np.int32), name='Const_0'),
                ],
                nodes=[
                    helper.make_node('Cast', inputs=['input_0'], outputs=['while_cond_158_while_Less__13_0'], name='while_cond_158_while_Less__13', domain='', to=TensorProto.FLOAT),
                    helper.make_node('Less', inputs=['while_cond_158_while_Less__13_0', 'const_fold_opt__17'], outputs=['while_cond_158_while_Less_0'], name='while_cond_158_while_Less', domain=''),
                    helper.make_node('Squeeze', inputs=['while_cond_158_while_Less_0'], outputs=['while_cond_158_while_Squeeze_0'], name='while_cond_158_while_Squeeze', domain=''),
                    helper.make_node(
                        'Loop',
                        inputs=['while_maximum_iterations_0', 'while_cond_158_while_Squeeze_0', 'input_0', 'Const_0'],
                        outputs=['while_loop_0', 'while_loop_1'],
                        name='while_loop',
                        body=helper.make_graph(
                            name='while_body',
                            inputs=[
                                helper.make_tensor_value_info('while_while_loop_counter_0', TensorProto.INT64, shape=[]),
                                helper.make_tensor_value_info('cond__15_0', TensorProto.BOOL, shape=[]),
                                helper.make_tensor_value_info('while_placeholder_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('while_add_const_0_0', TensorProto.INT32, shape=[1]),
                            ],
                            outputs=[
                                helper.make_tensor_value_info('cond___while_Identity_graph_outputs_Identity__3_0', TensorProto.BOOL, shape=[]),
                                helper.make_tensor_value_info('while_Identity_2_0', TensorProto.INT32, shape=[1]),
                                helper.make_tensor_value_info('while_add_const_0_0', TensorProto.INT32, shape=[1]),
                            ],
                            # Cannot use the same name as both a subgraph initializer and subgraph input: while_while_loop_counter_0
                            initializer=[numpy_helper.from_array(np.array(10, dtype=np.int64), name='while_while_loop_counter_0')],
                            nodes=[
                                helper.make_node('Add', inputs=['while_placeholder_0', 'while_add_const_0_0'], outputs=['while_Identity_2_0'], name='while_Add'),
                                helper.make_node('Cast', inputs=['while_Identity_2_0'], outputs=['cond___while_Less__13_0'], name='cond___while_Less__13', domain='', to=TensorProto.FLOAT),
                                helper.make_node('Less', inputs=['cond___while_Less__13_0', 'while_while_loop_counter_0'], outputs=['cond___while_Less_0'], name='cond___while_Less', domain=''),
                                helper.make_node('Squeeze', inputs=['cond___while_Less_0'], outputs=['cond___while_Identity_graph_outputs_Identity__3_0'], name='cond___while_Squeeze', domain=''),
                            ],
                        ),
                    ),
                    helper.make_node('Unsqueeze', inputs=['while_loop_0'], outputs=['Reshape_tensor_0'], name='Reshape_tensor', axes=[0]),
                    helper.make_node('Reshape', inputs=['Reshape_tensor_0', 'const_fold_opt__18'], outputs=['output_0'], name='Reshape'),
                ],
            ),
        )
        self.assertRaises(shape_inference.InferenceError, checker.check_model, model, True)


if __name__ == '__main__':
    unittest.main()



import unittest

import onnx
from onnx import checker, parser, utils


class TestFunction(unittest.TestCase):
    def _verify_function_set(self, extracted_model, function_set, func_domain):  # type: ignore
        checker.check_model(extracted_model)
        self.assertEqual(len(extracted_model.functions), len(function_set))
        for function in function_set:
            self.assertIsNotNone(
                next((f for f in extracted_model.functions
                if f.name == function and f.domain == func_domain), None))

    def test_extract_model_with_local_function(self) -> None:
        r'''
        #   1. build a model with graph below. extract models with output combinations
        #   2. validate extracted models' local functions
        #
        # model graph:
        #      i0                    i1                 i2
        #      |   __________________|__________________/_________
        #      |  |                  |             |   /          |
        #      |  |                  |             |  /           |
        #   func_add        func_identity          add         identity
        #    |  ___\___________\____________________|_________    |
        #    | |    \           \                   |  _______|___|
        #    | |     \           \                  | |       |   |
        #    add     function_nested_identity_add   add     function_nested_identity_add
        #     |                 |                    |              |
        #     |                 |                    |              |
        #   o_func_add      o_all_func0           o_no_func     o_all_func1
        #
        # where function_nested_identity_add is a function that is defined with functions:
        #       a               b
        #       |               |
        #   func_identity   func_identity
        #             \       /
        #             func_add
        #                |
        #                c
        #
        '''

        # function common
        func_domain = 'local'
        func_opset_imports = [onnx.helper.make_opsetid("", 14)]
        func_nested_opset_imports = [
            onnx.helper.make_opsetid("", 14), onnx.helper.make_opsetid(func_domain, 1)]

        # add function
        func_add_name = 'func_add'
        func_add_inputs = ['a', 'b']
        func_add_outputs = ['c']
        func_add_nodes = [onnx.helper.make_node('Add', ['a', 'b'], ['c'])]
        func_add = onnx.helper.make_function(
            func_domain,
            func_add_name,
            func_add_inputs,
            func_add_outputs,
            func_add_nodes,
            func_opset_imports)

        # identity function
        func_identity_name = 'func_identity'
        func_identity_inputs = ['a']
        func_identity_outputs = ['b']
        func_identity_nodes = [onnx.helper.make_node('Identity', ['a'], ['b'])]
        func_identity = onnx.helper.make_function(
            func_domain,
            func_identity_name,
            func_identity_inputs,
            func_identity_outputs,
            func_identity_nodes,
            func_opset_imports)

        # nested identity/add function
        func_nested_identity_add_name = 'func_nested_identity_add'
        func_nested_identity_add_inputs = ['a', 'b']
        func_nested_identity_add_outputs = ['c']
        func_nested_identity_add_nodes = [
            onnx.helper.make_node('func_identity', ['a'], ['a1'], domain=func_domain),
            onnx.helper.make_node('func_identity', ['b'], ['b1'], domain=func_domain),
            onnx.helper.make_node('func_add', ['a1', 'b1'], ['c'], domain=func_domain)]
        func_nested_identity_add = onnx.helper.make_function(
            func_domain,
            func_nested_identity_add_name,
            func_nested_identity_add_inputs,
            func_nested_identity_add_outputs,
            func_nested_identity_add_nodes,
            func_nested_opset_imports)

        # create graph nodes
        node_func_add = onnx.helper.make_node(func_add_name, ['i0', 'i1'], ['t0'], domain=func_domain)
        node_add0 = onnx.helper.make_node('Add', ['i1', 'i2'], ['t2'])
        node_add1 = onnx.helper.make_node('Add', ['t0', 't2'], ['o_func_add'])
        node_func_identity = onnx.helper.make_node(func_identity_name, ['i1'], ['t1'], domain=func_domain)
        node_identity = onnx.helper.make_node('Identity', ['i1'], ['t3'])
        node_add2 = onnx.helper.make_node('Add', ['t3', 't2'], ['o_no_func'])
        node_func_nested0 = onnx.helper.make_node(
            func_nested_identity_add_name,
            ['t0', 't1'],
            ['o_all_func0'],
            domain=func_domain)
        node_func_nested1 = onnx.helper.make_node(
            func_nested_identity_add_name,
            ['t3', 't2'],
            ['o_all_func1'],
            domain=func_domain)

        graph_name = 'graph_with_imbedded_functions'
        ir_version = 8
        opset_imports = [onnx.helper.make_opsetid("", 14), onnx.helper.make_opsetid("local", 1)]
        tensor_type_proto = onnx.helper.make_tensor_type_proto(elem_type=2, shape=[5])

        graph = onnx.helper.make_graph(
            [node_func_add, node_add0, node_add1, node_func_identity, node_identity,
            node_func_nested0, node_func_nested1, node_add2],
            graph_name,
            [
                onnx.helper.make_value_info(name='i0', type_proto=tensor_type_proto),
                onnx.helper.make_value_info(name='i1', type_proto=tensor_type_proto),
                onnx.helper.make_value_info(name='i2', type_proto=tensor_type_proto)],
            [
                onnx.helper.make_value_info(name='o_no_func', type_proto=tensor_type_proto),
                onnx.helper.make_value_info(name='o_func_add', type_proto=tensor_type_proto),
                onnx.helper.make_value_info(name='o_all_func0', type_proto=tensor_type_proto),
                onnx.helper.make_value_info(name='o_all_func1', type_proto=tensor_type_proto)])

        meta = {
            'ir_version': ir_version,
            'opset_imports': opset_imports,
            'producer_name': 'test_extract_model_with_local_function',
            'functions': [func_identity, func_add, func_nested_identity_add],
        }
        model = onnx.helper.make_model(graph, **meta)

        checker.check_model(model)
        extracted_with_no_funcion = utils.Extractor(model).extract_model(['i0', 'i1', 'i2'], ['o_no_func'])
        self._verify_function_set(extracted_with_no_funcion, {}, func_domain)

        extracted_with_add_funcion = utils.Extractor(model).extract_model(['i0', 'i1', 'i2'], ['o_func_add'])
        self._verify_function_set(extracted_with_add_funcion, {func_add_name}, func_domain)

        extracted_with_o_all_funcion0 = utils.Extractor(model).extract_model(['i0', 'i1', 'i2'], ['o_all_func0'])
        self._verify_function_set(
            extracted_with_o_all_funcion0,
            {func_add_name, func_identity_name, func_nested_identity_add_name},
            func_domain)

        extracted_with_o_all_funcion1 = utils.Extractor(model).extract_model(['i0', 'i1', 'i2'], ['o_all_func1'])
        self._verify_function_set(
            extracted_with_o_all_funcion1,
            {func_add_name, func_identity_name, func_nested_identity_add_name},
            func_domain)

        extracted_with_o_all_funcion2 = utils.Extractor(model).extract_model(
            ['i0', 'i1', 'i2'], ['o_no_func', 'o_func_add', 'o_all_func0', 'o_all_func1'])
        self._verify_function_set(
            extracted_with_o_all_funcion2,
            {func_add_name, func_identity_name, func_nested_identity_add_name},
            func_domain)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
import unittest

from onnx import defs, checker, helper


class TestRelu(unittest.TestCase):

    def test_elu(self) -> None:
        self.assertTrue(defs.has('Elu'))
        node_def = helper.make_node(
            'Elu', ['X'], ['Y'], alpha=1.0)
        checker.check_node(node_def)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

import numpy as np  # type: ignore

import onnx
from ..base import Base
from . import expect


class Mish(Base):

    @staticmethod
    def export() -> None:
        node = onnx.helper.make_node(
            'Mish',
            inputs=['X'],
            outputs=['Y']
        )

        input_data = np.linspace(-10, 10, 10000, dtype=np.float32)

        # Calculate expected output data
        expected_output = input_data * np.tanh(np.log1p(np.exp(input_data)))

        expect(node, inputs=[input_data], outputs=[expected_output],
               name='test_mish')



# SPDX-License-Identifier: Apache-2.0
import unittest
import onnx.hub as hub
from onnx import ModelProto
import glob
from os.path import join
import pytest  # type: ignore
import os


@pytest.mark.skipif(
    'TEST_HUB' not in os.environ or not os.environ['TEST_HUB'],
    reason="Conserving Git LFS quota"
)
class TestModelHub(unittest.TestCase):
    def setUp(self) -> None:
        self.name = "MNIST"
        self.repo = "onnx/models:main"
        self.opset = 7

    def test_force_reload(self) -> None:
        model = hub.load(self.name, self.repo, force_reload=True)
        self.assertIsInstance(model, ModelProto)

        cached_files = list(glob.glob(join(hub.get_dir(), "**", "*.onnx"), recursive=True))
        self.assertGreaterEqual(len(cached_files), 1)

    def test_listing_models(self) -> None:
        model_info_list_1 = hub.list_models(self.repo, model="mnist", tags=["vision"])
        model_info_list_2 = hub.list_models(self.repo, tags=["vision"])
        model_info_list_3 = hub.list_models(self.repo)

        self.assertGreater(len(model_info_list_1), 1)
        self.assertGreater(len(model_info_list_2), len(model_info_list_1))
        self.assertGreater(len(model_info_list_3), len(model_info_list_2))

    def test_basic_usage(self) -> None:
        model = hub.load(self.name, self.repo)
        self.assertIsInstance(model, ModelProto)

        cached_files = list(glob.glob(join(hub.get_dir(), "**", "*.onnx"), recursive=True))
        self.assertGreaterEqual(len(cached_files), 1)

    def test_custom_cache(self) -> None:
        old_cache = hub.get_dir()
        new_cache = join(old_cache, "custom")
        hub.set_dir(new_cache)

        model = hub.load(self.name, self.repo)
        self.assertIsInstance(model, ModelProto)

        cached_files = list(glob.glob(join(new_cache, "**", "*.onnx"), recursive=True))
        self.assertGreaterEqual(len(cached_files), 1)

        hub.set_dir(old_cache)

    def test_download_with_opset(self) -> None:
        model = hub.load(self.name, self.repo, opset=8)
        self.assertIsInstance(model, ModelProto)

    def test_opset_error(self) -> None:
        self.assertRaises(AssertionError, lambda: hub.load(self.name, self.repo, opset=-1))

    def test_manifest_not_found(self) -> None:
        self.assertRaises(AssertionError, lambda: hub.load(self.name, "onnx/models:unknown", silent=True))

    def test_verify_repo_ref(self) -> None:
        # Not trusted repo:
        verified = hub._verify_repo_ref("mhamilton723/models")
        self.assertFalse(verified)

        # Not trusted repo:
        verified = hub._verify_repo_ref("onnx/models:unknown")
        self.assertFalse(verified)

        # Trusted repo:
        verified = hub._verify_repo_ref(self.repo)
        self.assertTrue(verified)

    def test_get_model_info(self) -> None:
        hub.get_model_info("mnist", self.repo, opset=8)
        hub.get_model_info("mnist", self.repo)
        self.assertRaises(AssertionError, lambda: hub.get_model_info("mnist", self.repo, opset=-1))



from typing import Dict, Tuple, Sequence, Optional, Set, overload
from onnx import AttributeProto

class SchemaError(Exception):
    ...

class OpSchema:
    @property
    def file(self) -> str: ...
    @property
    def line(self) -> int: ...
    @property
    def support_level(self) -> SupportType: ...
    @property
    def doc(self) -> Optional[str]: ...
    @property
    def since_version(self) -> int: ...
    @property
    def deprecated(self) -> bool: ...
    @property
    def domain(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def min_input(self) -> int: ...
    @property
    def max_input(self) -> int: ...
    @property
    def min_output(self) -> int: ...
    @property
    def max_output(self) -> int: ...
    @property
    def attributes(self) -> Dict[str, Attribute]: ...
    @property
    def inputs(self) -> Sequence[FormalParameter]: ...
    @property
    def outputs(self) -> Sequence[FormalParameter]: ...
    @property
    def type_constraints(self) -> Sequence[TypeConstraintParam]: ...
    @property
    def has_type_and_shape_inference_function(self) -> bool: ...
    @property
    def has_data_propagation_function(self) -> bool: ...

    @staticmethod
    def is_infinite(v: int) -> bool: ...

    def consumed(self, schema: OpSchema, i: int) -> Tuple[UseType, int]: ...

    class TypeConstraintParam:
        @property
        def type_param_str(self) -> str: ...
        @property
        def description(self) -> str: ...
        @property
        def allowed_type_strs(self) -> Sequence[str]: ...

    class FormalParameterOption:
        Single: 'OpSchema.FormalParameterOption' = ...
        Optional: 'OpSchema.FormalParameterOption' = ...
        Variadic: 'OpSchema.FormalParameterOption' = ...

    class DifferentiationCategory:
        Unknown: 'OpSchema.DifferentiationCategory' = ...
        Differentiable: 'OpSchema.DifferentiationCategory' = ...
        NonDifferentiable: 'OpSchema.DifferentiationCategory' = ...

    class FormalParameter:
        @property
        def name(self) -> str: ...
        @property
        def types(self) -> Set[str]: ...
        @property
        def typeStr(self) -> str: ...
        @property
        def description(self) -> str: ...
        @property
        def option(self) -> OpSchema.FormalParameterOption: ...
        @property
        def isHomogeneous(self) -> bool: ...
        @property
        def differentiationCategory(self) -> OpSchema.DifferentiationCategory: ...

    class AttrType:
        FLOAT: 'OpSchema.AttrType' = ...
        INT: 'OpSchema.AttrType' = ...
        STRING: 'OpSchema.AttrType' = ...
        TENSOR: 'OpSchema.AttrType' = ...
        GRAPH: 'OpSchema.AttrType' = ...
        SPARSE_TENSOR: 'OpSchema.AttrType' = ...
        TYPE_PROTO: 'OpSchema.AttrType' = ...
        FLOATS: 'OpSchema.AttrType' = ...
        INTS: 'OpSchema.AttrType' = ...
        STRINGS: 'OpSchema.AttrType' = ...
        TENSORS: 'OpSchema.AttrType' = ...
        GRAPHS: 'OpSchema.AttrType' = ...
        SPARSE_TENSORS: 'OpSchema.AttrType' = ...
        TYPE_PROTOS: 'OpSchema.AttrType' = ...

    class Attribute:
        @property
        def name(self) -> str: ...
        @property
        def description(self) -> str: ...
        @property
        def type(self) -> OpSchema.AttrType: ...
        @property
        def default_value(self) -> AttributeProto: ...
        @property
        def required(self) -> bool: ...


    class SupportType(int):
        COMMON: 'OpSchema.SupportType' = ...
        EXPERIMENTAL: 'OpSchema.SupportType' = ...


    class UseType:
        DEFAULT: 'OpSchema.UseType' = ...
        CONSUME_ALLOWED: 'OpSchema.UseType' = ...
        CONSUME_ENFORCED: 'OpSchema.UseType' = ...


def has_schema(op_type: str) -> bool: ...
def schema_version_map() -> Dict[str, Tuple[int, int]]: ...
@overload
def get_schema(op_type: str, max_inclusive_version: int, domain: str = "") -> OpSchema: ...
@overload
def get_schema(op_type: str, domain: str = "") -> OpSchema: ...
def get_all_schemas() -> Sequence[OpSchema]: ...
def get_all_schemas_with_history() -> Sequence[OpSchema]: ...



# SPDX-License-Identifier: Apache-2.0

import unittest
import onnx
from onnx.tools import update_model_dims
from onnx import helper, TensorProto


class TestToolsFunctions(unittest.TestCase):
    def test_update_inputs_outputs_dim(self) -> None:
        node_def = helper.make_node(
            "Conv",
            inputs=['x', 'W'],
            outputs=['y'],
            kernel_shape=[3, 3],
            strides=[2, 2],
        )
        graph_def = helper.make_graph(
            [node_def],
            'test',
            [helper.make_tensor_value_info('x', TensorProto.FLOAT, [1, 1, 5, 5]),
             helper.make_tensor_value_info('W', TensorProto.FLOAT, [1, 1, 3, 3])],
            [helper.make_tensor_value_info('y', TensorProto.FLOAT, [1, 1, 2, 2])]
        )
        model_def = helper.make_model(graph_def, producer_name='test')
        updated_def = update_model_dims.update_inputs_outputs_dims(
            model_def,
            {
                "x": [1, 1, 'x1', -1],
                "W": [1, 1, 3, 3],
            },
            {
                "y": [1, 1, -1, -1],
            })
        onnx.checker.check_model(updated_def)
        self.assertEqual(updated_def.graph.input[0].type.tensor_type.shape.dim[2].dim_param, 'x1')
        self.assertEqual(updated_def.graph.input[0].type.tensor_type.shape.dim[3].dim_param, 'x_3')
        self.assertEqual(updated_def.graph.output[0].type.tensor_type.shape.dim[2].dim_param, 'y_2')
        self.assertEqual(updated_def.graph.output[0].type.tensor_type.shape.dim[3].dim_param, 'y_3')


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from onnx import AttributeProto, NodeProto, GraphProto, ModelProto, TensorProto, IR_VERSION

import io
import onnx
import os
import tempfile
import unittest

from onnx import helper


class TestBasicFunctions(unittest.TestCase):

    def _simple_model(self) -> ModelProto:
        # Create a ModelProto.
        model = ModelProto()
        model.ir_version = IR_VERSION
        return model

    def _simple_tensor(self) -> TensorProto:
        # Create a TensorProto.
        tensor = helper.make_tensor(
            name='test-tensor',
            data_type=TensorProto.FLOAT,
            dims=(2, 3, 4),
            vals=[x + 0.5 for x in range(24)]
        )
        return tensor

    def test_save_and_load_model(self) -> None:
        proto = self._simple_model()
        cls = ModelProto
        proto_string = onnx._serialize(proto)

        # Test if input is string
        loaded_proto = onnx.load_model_from_string(proto_string)
        self.assertTrue(proto == loaded_proto)

        # Test if input has a read function
        f = io.BytesIO()
        onnx.save_model(proto_string, f)
        f = io.BytesIO(f.getvalue())
        loaded_proto = onnx.load_model(f, cls)
        self.assertTrue(proto == loaded_proto)

        # Test if input is a file name
        try:
            fi = tempfile.NamedTemporaryFile(delete=False)
            onnx.save_model(proto, fi)
            fi.close()

            loaded_proto = onnx.load_model(fi.name, cls)
            self.assertTrue(proto == loaded_proto)
        finally:
            os.remove(fi.name)

    def test_save_and_load_tensor(self) -> None:
        proto = self._simple_tensor()
        cls = TensorProto
        proto_string = onnx._serialize(proto)

        # Test if input is string
        loaded_proto = onnx.load_tensor_from_string(proto_string)
        self.assertTrue(proto == loaded_proto)

        # Test if input has a read function
        f = io.BytesIO()
        onnx.save_tensor(loaded_proto, f)
        f = io.BytesIO(f.getvalue())
        loaded_proto = onnx.load_tensor(f, cls)
        self.assertTrue(proto == loaded_proto)

        # Test if input is a file name
        try:
            tfile = tempfile.NamedTemporaryFile(delete=False)
            onnx.save_tensor(proto, tfile)
            tfile.close()

            loaded_proto = onnx.load_tensor(tfile.name, cls)
            self.assertTrue(proto == loaded_proto)
        finally:
            os.remove(tfile.name)

    def test_existence(self) -> None:
        try:
            AttributeProto
            NodeProto
            GraphProto
            ModelProto
        except Exception as e:
            self.fail(
                f'Did not find proper onnx protobufs. Error is: {e}')

    def test_version_exists(self) -> None:
        model = ModelProto()
        # When we create it, graph should not have a version string.
        self.assertFalse(model.HasField('ir_version'))
        # We should touch the version so it is annotated with the current
        # ir version of the running ONNX
        model.ir_version = IR_VERSION
        model_string = model.SerializeToString()
        model.ParseFromString(model_string)
        self.assertTrue(model.HasField('ir_version'))
        # Check if the version is correct.
        self.assertEqual(model.ir_version, IR_VERSION)


if __name__ == '__main__':
    unittest.main()
